# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.svm._classes import LinearSVCSklearn
from ...utils.multiclass import check_classification_targets
from pai4sk.svm._base import _get_liblinear_solver_type
import warnings
from ...exceptions import NotFittedError
import numpy as np
from .utils import _check_snapml_types, _check_snapml_type, _is_gpu_matrix_type, _get_data_size, _check_uparams
from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type
_mpi_enabled = _is_mpi_enabled()


class LinearSVC(LinearSVCSklearn):
    """Linear Support Vector Classification.

    Similar to SVC with parameter kernel='linear', but implemented in terms of
    liblinear rather than libsvm, so it has more flexibility in the choice of
    penalties and loss functions and should scale better to large numbers of
    samples.

    This class supports both dense and sparse input and the multiclass support
    is handled according to a one-vs-the-rest scheme.

    Read more in the :ref:`User Guide <svm_classification>`.

    For SnapML solver this supports both local and distributed(MPI) method of execution.

    Parameters
    ----------
    penalty : string, 'l1' or 'l2' (default='l2')
        Specifies the norm used in the penalization. The 'l2'
        penalty is the standard used in SVC. The 'l1' leads to ``coef_``
        vectors that are sparse.

    loss : string, 'hinge' or 'squared_hinge' (default='squared_hinge')
        Specifies the loss function. 'hinge' is the standard SVM loss
        (used e.g. by the SVC class) while 'squared_hinge' is the
        square of the hinge loss.

    dual : bool, (default=True)
        Select the algorithm to either solve the dual or primal
        optimization problem. Prefer dual=False when n_samples > n_features.

    tol : float, optional (default=1e-4)
        Tolerance for stopping criteria.

    C : float, optional (default=1.0)
        Penalty parameter C of the error term.

    multi_class : string, 'ovr' or 'crammer_singer' (default='ovr')
        Determines the multi-class strategy if `y` contains more than
        two classes.
        ``"ovr"`` trains n_classes one-vs-rest classifiers, while
        ``"crammer_singer"`` optimizes a joint objective over all classes.
        While `crammer_singer` is interesting from a theoretical perspective
        as it is consistent, it is seldom used in practice as it rarely leads
        to better accuracy and is more expensive to compute.
        If ``"crammer_singer"`` is chosen, the options loss, penalty and dual
        will be ignored.

    fit_intercept : boolean, optional (default=True)
        Whether to calculate the intercept for this model. If set
        to false, no intercept will be used in calculations
        (i.e. data is expected to be already centered).

    intercept_scaling : float, optional (default=1)
        When self.fit_intercept is True, instance vector x becomes
        ``[x, self.intercept_scaling]``,
        i.e. a "synthetic" feature with constant value equals to
        intercept_scaling is appended to the instance vector.
        The intercept becomes intercept_scaling * synthetic feature weight
        Note! the synthetic feature weight is subject to l1/l2 regularization
        as all other features.
        To lessen the effect of regularization on synthetic feature weight
        (and therefore on the intercept) intercept_scaling has to be increased.

    class_weight : {dict, 'balanced'}, optional
        Set the parameter C of class i to ``class_weight[i]*C`` for
        SVC. If not given, all classes are supposed to have
        weight one.
        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

    verbose : int, (default=0)
        Enable verbose output. Note that this setting takes advantage of a
        per-process runtime setting in liblinear that, if enabled, may not work
        properly in a multithreaded context.

    random_state : int, RandomState instance or None, optional (default=None)
        The seed of the pseudo random number generator to use when shuffling
        the data for the dual coordinate descent (if ``dual=True``). When
        ``dual=False`` the underlying implementation of :class:`LinearSVC`
        is not random and ``random_state`` has no effect on the results. If
        int, random_state is the seed used by the random number generator; If
        RandomState instance, random_state is the random number generator; If
        None, the random number generator is the RandomState instance used by
        `np.random`.

    max_iter : int, (default=1000)
        The maximum number of iterations to be run.

    use_gpu : bool, default : True
        Flag for indicating the hardware platform used for training. If True, the training
        is performed using the GPU. If False, the training is performed using the CPU.
        The value of this parameter is subjected to changed based on the training data unless set explicitly.
        Applicable only for snapml solver

    device_ids : array-like of int, default : []
        If use_gpu is True, it indicates the IDs of the GPUs used for training.
        For single GPU training, set device_ids to the GPU ID to be used for training, 
        e.g., [0]. For multi-GPU training, set device_ids to a list of GPU IDs to be used 
        for training, e.g., [0, 1]. 
        Applicable only for snapml solver

    num_threads : int, default : 1
        The number of threads used for running the training. The value of this parameter
        should be a multiple of 32 if the training is performed on GPU (use_gpu=True)
        (default value for GPU is 256). Applicable only for snapml solver

    return_training_history : str or None, default : None
        How much information about the training should be collected and returned by the fit function. By
        default no information is returned (None), but this parameter can be set to "summary", to obtain
        summary statistics at the end of training, or "full" to obtain a complete set of statistics
        for the entire training procedure. Note, enabling either option will result in slower training.
        Applicable only for snapml solver

    Attributes
    ----------
    coef_ : array, shape = [n_features] if n_classes == 2 else [n_classes, n_features]
        Weights assigned to the features (coefficients in the primal
        problem). This is only available in the case of a linear kernel.

        ``coef_`` is a readonly property derived from ``raw_coef_`` that
        follows the internal memory layout of liblinear.

    intercept_ : array, shape = [1] if n_classes == 2 else [n_classes]
        Constants in decision function.

    training_history_ : dict
        It returns a dictionary with the following keys : 'epochs', 't_elap_sec', 'train_obj'.
        If 'return_training_history' is set to "summary", 'epochs' contains the total number of
        epochs performed, 't_elap_sec' contains the total time for completing all of those epochs.
        If 'return_training_history' is set to "full", 'epochs' indicates the number of epochs
        that have elapsed so far, and 't_elap_sec' contains the time to do those epochs.
        'train_obj' is the training loss.
        Applicable only for snapml solver.

    support_ : array-like,  shape (n_SV)
        indices of the support vectors.

    n_support_ : int
        Number of support vectors.

    n_iter_ : array, shape (n_classes,) or (1, )
        Actual number of iterations for all classes to reach the specified tolerance. 
        If binary or multinomial, it returns only 1 element.

    Examples
    --------
    >>> from pai4sk.svm import LinearSVC
    >>> from pai4sk.datasets import make_classification
    >>> X, y = make_classification(n_features=4, random_state=0)
    >>> clf = LinearSVC(random_state=0, tol=1e-5)
    >>> clf.fit(X, y)
    LinearSVC(C=1.0, class_weight=None, dual=True, fit_intercept=True,
         intercept_scaling=1, loss='squared_hinge', max_iter=1000,
         multi_class='ovr', penalty='l2', random_state=0, tol=1e-05, verbose=0)
    >>> print(clf.coef_)
    [[0.085... 0.394... 0.498... 0.375...]]
    >>> print(clf.intercept_)
    [0.284...]
    >>> print(clf.predict([[0, 0, 0, 0]]))
    [1]

    Notes
    -----
    The underlying C implementation uses a random number generator to
    select features when fitting the model. It is thus not uncommon
    to have slightly different results for the same input data. If
    that happens, try with a smaller ``tol`` parameter.

    The underlying implementation, liblinear, uses a sparse internal
    representation for the data that will incur a memory copy.

    Predict output may not match that of standalone liblinear in certain
    cases. See :ref:`differences from liblinear <liblinear_differences>`
    in the narrative documentation.

    References
    ----------
    `LIBLINEAR: A Library for Large Linear Classification
    <http://www.csie.ntu.edu.tw/~cjlin/liblinear/>`__

    See also
    --------
    SVC
        Implementation of Support Vector Machine classifier using libsvm:
        the kernel can be non-linear but its SMO algorithm does not
        scale to large number of samples as LinearSVC does.

        Furthermore SVC multi-class mode is implemented using one
        vs one scheme while LinearSVC uses one vs the rest. It is
        possible to implement one vs the rest with SVC by using the
        :class:`pai4sk.multiclass.OneVsRestClassifier` wrapper.

        Finally SVC can fit dense data without memory copy if the input
        is C-contiguous. Sparse data will still incur memory copy though.

    pai4sk.linear_model.SGDClassifier
        SGDClassifier can optimize the same cost function as LinearSVC
        by adjusting the penalty and loss parameters. In addition it requires
        less memory, allows incremental (online) learning, and implements
        various loss functions and regularization regimes.

    """
    __slots__ = ('_LinearSVCSklearn')
    
    _param_list = []    # Store original arguments
    _uparams = None     # Store user provided arguments

    # Override the existing function to include the current list of parameters
    # as the init signature is modified by the decorator
    def _get_param_names(self):
        return self._param_list

    @_check_uparams
    def __init__(self, penalty='l2', loss='squared_hinge', dual=True, tol=1e-4,
                 C=1.0, multi_class='ovr', fit_intercept=True,
                 intercept_scaling=1, class_weight=None, verbose=0,
                 random_state=None, max_iter=1000,
                 use_gpu=True,                      # pai4sk
                 device_ids=[],                     # pai4sk
                 num_threads=1,                     # pai4sk
                 return_training_history=None):     # pai4sk

        self.dual = dual
        self.tol = tol
        self.C = C
        self.multi_class = multi_class
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.verbose = verbose
        self.random_state = random_state
        self.max_iter = max_iter
        self.penalty = penalty
        self.loss = loss
        self.use_gpu = use_gpu
        self.device_ids = device_ids
        self.num_threads = num_threads
        self.return_training_history = return_training_history
        self._snapsvm = None
        self._LinearSVCSklearn = super(LinearSVC, self)

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._LinearSVCSklearn = super(LinearSVC, self)

    # Validate inputs and initialize pai4sk
    def _check_set_snapml(self, X, y):
        snap_args = {}
        mod_params = []
        unsupported_params = []

        if ( not self.random_state is None):
            unsupported_params.append("random_state")

        if (not self.dual):
            unsupported_params.append("dual")

        if (type(self.class_weight) is dict):
            unsupported_params.append("class_weight")

        if (not self.multi_class == 'ovr'):
            unsupported_params.append("multi_class")

        if ('loss' in self._uparams and self.loss != 'hinge'):
            unsupported_params.append("multi_class")
        else:
            mod_params.append("loss")
            self.loss = 'hinge'

        # choose CPU implementation for smaller dataset (in-memory size < 400kb)
        if not _is_data_snapml_partition_type(X) and not _is_gpu_matrix_type(X, y):
            if _get_data_size(X) < 400 and not 'use_gpu' in self._uparams :
                self.use_gpu = False
                mod_params.append("use_gpu")

        if not _check_snapml_types(X, y):
            unsupported_params.append("X/y")
        else:
            if not _is_data_snapml_partition_type(y) and not _is_gpu_matrix_type(X, y):
                uniq_labs_ = np.unique(y)
                if (len(uniq_labs_) == 2):
                    if not((uniq_labs_[0] == -1 and uniq_labs_[1] == 1) or 
                        (uniq_labs_[0] == -1 and uniq_labs_[1] == 1) or 
                        (uniq_labs_[0] == 0 and uniq_labs_[1] == 1) or 
                        (uniq_labs_[0] == 1 and uniq_labs_[1] == 0)):
                        unsupported_params.append("y")
                else:
                    try:
                        uniq_labs_.astype(np.float32)
                    except ValueError:
                        unsupported_params.append("y")

        if not unsupported_params:
            # Check for GPUs
            snap_args['use_gpu'] = False if self.use_gpu is None else self.use_gpu
            snap_args['device_ids'] = [] if self.device_ids is None else self.device_ids
            if self.use_gpu:
                from pai4sk import sml_utils
                ndevices = sml_utils.get_gpu_count()
                if ndevices == 0:
                    if 'use_gpu' in self._uparams:
                        raise ValueError("SnapML: GPU device not found")
                    mod_params.append("use_gpu")
                    self.use_gpu = False

            if 'max_iter' in self._uparams:
                snap_args['max_iter'] = self.max_iter
            else:
                mod_params.append("max_iter")
                snap_args['max_iter'] = 1000

            if 'tol' in self._uparams:
                snap_args['tol'] = self.tol
            else:
                mod_params.append("tol")
                snap_args['tol'] = 0.001

            if 'fit_intercept' in self._uparams:
                snap_args['fit_intercept'] = self.fit_intercept
            else:
                mod_params.append("fit_intercept")
                self.fit_intercept = False

            snap_args['verbose'] = False if self.verbose == 0 else True
            snap_args['regularizer'] = 1 if self.C is None else 1/self.C
            snap_args['intercept_scaling'] = 1.0 if self.intercept_scaling is None else self.intercept_scaling
            snap_args['return_training_history'] = self.return_training_history
            snap_args['class_weight'] = self.class_weight
            snap_args['num_threads'] = self.num_threads

            from pai4sk import SupportVectorMachine as _snapsvm
            self._snapsvm = _snapsvm(**snap_args)

            warnings.warn("SnapML: Default solver is set to 'snapml'.")
                
        if unsupported_params:
            warnMsg = "SnapML: fallback to sklearn implementation due to unsupported parameters/values: " + ", ".join([str(x) for x in unsupported_params])
            warnings.warn(warnMsg, UserWarning)
            return False

        if mod_params:
            warnMsg = "SnapML: Default values for these parameters are modified for 'snapml' solver: " + ", ".join([str(x) for x in mod_params])
            warnings.warn(warnMsg, UserWarning)

        return True

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.
            For SnapML solver it also supports input of types SnapML data partition and DeviceNDArray.
        y : array-like, shape = [n_samples]
            Target vector relative to X
        sample_weight : array-like, shape = [n_samples], optional
            Array of weights that are assigned to individual
            samples. If not provided,
            then each sample is given unit weight.
        Returns
        -------
        self : object
        """
        # FIXME Remove l1/l2 support in 1.0 -----------------------------------
        msg = ("loss='%s' has been deprecated in favor of "
               "loss='%s' as of 0.16. Backward compatibility"
               " for the loss='%s' will be removed in %s")

        if self.loss in ('l1', 'l2'):
            old_loss = self.loss
            self.loss = {'l1': 'hinge', 'l2': 'squared_hinge'}.get(self.loss)
            warnings.warn(msg % (old_loss, self.loss, old_loss, '1.0'),
                          DeprecationWarning)
        # ---------------------------------------------------------------------

        if self.C < 0:
            raise ValueError("Penalty term must be positive; got (C=%r)"
                             % self.C)

        _get_liblinear_solver_type(self.multi_class, self.penalty, self.loss, self.dual)

        if self._check_set_snapml(X, y):
            # snap-ml fit
            self.training_history_ = self._snapsvm.fit(X, y)

            if not _mpi_enabled:
                # Attributes only supported with snap-ml-local
                self.support_ = self._snapsvm.support_
                self.n_support_ = self._snapsvm.n_support_
                self.n_iter_ = self._snapsvm.n_iter_

            self.classes_ = self._snapsvm.uniq_labs_.astype(np.int)
            self.coef_ = self._snapsvm.model_
            if (self.fit_intercept):
                self.coef_ = self.coef_[:,:-1]
                self.intercept_ = self.coef_[:,-1]
            else:
                #self.coef_ is already of required shape
                self.intercept_ = np.empty(shape=())

            return self
        else:
            lr_sk = self._LinearSVCSklearn.fit(X, y, sample_weight)
            self.classes_ = lr_sk.classes_
            self.coef_ = lr_sk.coef_
            self.intercept_ = lr_sk.intercept_
            self.n_iter_ = lr_sk.n_iter_
            self.training_history_ = []
            return self
            

    def predict(self, X, num_threads=0):

        """
        Class predictions
        The returned class estimates.
        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting class estimates.
            For SnapML solver it also supports input of type SnapML data partition.
       num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.
        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """
        if not hasattr(self, "coef_"):
            raise NotFittedError("Call fit before prediction")

        if _check_snapml_type(X) and self._snapsvm:
            return self._snapsvm.predict(X, num_threads)
        else:
            return self._LinearSVCSklearn.predict(X)

    def decision_function(self, X, num_threads=0):
        """
        Predicts confidence scores.

        The confidence score of a sample is the signed distance of that sample to the decision boundary.

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting distances to the decision boundary.
            For SnapML solver it also supports input of type SnapML data partition.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        proba: array-like, shape = (n_samples,) or (n_sample, n_classes)
            Returns the distance to the decision boundary of the samples in X.
        """
        if not hasattr(self, "coef_"):
            raise NotFittedError("Call fit before prediction")

        if _check_snapml_type(X) and self._snapsvm:
            return self._snapsvm.decision_function(X, num_threads)
        else:
            return self._LinearSVCSklearn.decision_function(X)
